@@ -16,7 +16,7 @@ use minidump_processor::{
     http_symbol_supplier, simple_symbol_supplier, MultiSymbolProvider, ProcessorOptions, Symbolizer,
 };
 
-use clap::{crate_version, App, AppSettings, Arg};
+use clap::{crate_version, App, AppSettings, Arg, ArgGroup};
 use log::error;
 use simplelog::{
     ColorChoice, ConfigBuilder, Level, LevelFilter, TermLogger, TerminalMode, WriteLogger,
@@ -24,48 +24,95 @@ use simplelog::{
 
 #[cfg_attr(test, allow(dead_code))]
 fn main() {
-    let matches = App::new("minidump_stackwalk")
+    let matches = App::new("minidump-stackwalk")
         .version(crate_version!())
-        .about("Analyzes minidumps and produces a machine-readable JSON report.")
+        .about("Analyzes minidumps and produces a report (either human-readable or JSON).")
         .setting(AppSettings::NextLineHelp)
         .setting(AppSettings::ColoredHelp)
+        .setting(AppSettings::DeriveDisplayOrder)
+        .usage("minidump-stackwalk [FLAGS] [OPTIONS] <minidump> [--] [symbols-path]...")
+        .arg(
+            Arg::with_name("json")
+                .long("json")
+                .long_help("Emit a machine-readable JSON report (the default).
+
+The schema for this output is officially documented here:
+https://github.com/luser/rust-minidump/blob/master/minidump-processor/json-schema.md\n\n\n")
+        )
         .arg(
             Arg::with_name("human")
-                .long_help("Emit a human-readable report instead.
+                .long("human")
+                .long_help("Emit a human-readable report.
 
 The human-readable report does not have a specified format, and may not have as \
 many details as the default JSON format. It is intended for quickly inspecting \
 a crash or debugging rust-minidump itself.\n\n\n")
-                .long("human")
         )
         .arg(
-            Arg::with_name("pretty")
-                .help("Pretty-print JSON output.")
-                .long("pretty")
+            Arg::with_name("cyborg")
+                .long("cyborg")
+                .takes_value(true)
+                .long_help("Combine --human and --json
+
+Because this creates two output streams, you must specify a path to write the --json
+output to. The --human output will be the 'primary' output and default to stdout, which
+can be configured with --output-file as normal.\n\n\n")
+        )
+        .group(ArgGroup::with_name("output-format")
+            .args(&["json", "human", "cyborg"])
+        )
+        .arg(
+            Arg::with_name("output-file")
+                .long("output-file")
+                .takes_value(true)
+                .help("Where to write the output to (if unspecified, stdout is used)")
+        )
+        .arg(
+            Arg::with_name("log-file")
+                .long("log-file")
+                .takes_value(true)
+                .help("Where to write logs to (if unspecified, stderr is used)")
         )
         .arg(
             Arg::with_name("verbose")
-                .long_help("Set the level of verbosity (off, error (default), warn, info, debug, trace)
+                .long("verbose")
+                .possible_values(&["off", "error", "warn", "info", "debug", "trace"])
+                .default_value("error")
+                .takes_value(true)
+                .long_help("Set the logging level.
 
 The unwinder has been heavily instrumented with `trace` logging, so if you want to debug why \
 an unwind happened the way it did, --verbose=trace is very useful (all unwinder logging will \
 be prefixed with `unwind:`).\n\n\n")
-                .long("verbose")
-                .default_value("error")
-                .takes_value(true),
+        )
+        .arg(
+            Arg::with_name("pretty")
+                .long("pretty")
+                .help("Pretty-print --json output.")
+        )
+        .arg(
+            Arg::with_name("brief")
+                .long("brief")
+                .help("Provide a briefer --human report.
+
+Only provides the top-level summary and a backtrace of the crashing thread.")
         )
         .arg(
             Arg::with_name("raw-json")
+                .long("raw-json")
+                .takes_value(true)
                 .long_help("An input JSON file with the extra information.
 
 This is a gross hack for some legacy side-channel information that mozilla uses. It will \
 hopefully be phased out and deprecated in favour of just using custom streams in the \
 minidump itself.\n\n\n")
-                .long("raw-json")
-                .takes_value(true),
         )
         .arg(
             Arg::with_name("symbols-url")
+                .long("symbols-url")
+                .multiple(true)
+                .takes_value(true)
+                .number_of_values(1)
                 .long_help("base URL from which URLs to symbol files can be constructed.
 
 If multiple symbols-url values are provided, they will each be tried in order until \
@@ -77,12 +124,11 @@ symbol server protocol. For more details, see the Tecken docs:
 https://tecken.readthedocs.io/en/latest/
 
 Example symbols-url value: https://symbols.mozilla.org/\n\n\n")
-                .multiple(true)
-                .long("symbols-url")
-                .takes_value(true),
         )
         .arg(
             Arg::with_name("symbols-cache")
+                .long("symbols-cache")
+                .takes_value(true)
                 .long_help("A directory in which downloaded symbols can be stored.
                     
 Symbol files can be very large, so we recommend placing cached files in your \
@@ -91,13 +137,12 @@ To this end, the default value for this flag is a `rust-minidump-cache` \
 subdirectory of `std::env::temp_dir()` (usually /tmp/rust-minidump-cache on linux).
 
 symbols-cache must be on the same filesystem as symbols-tmp (if that doesn't mean anything to \
-you, don't worry about it, you're probably not doing something that will run afoul of it).
-\n\n\n")
-                .long("symbols-cache")
-                .takes_value(true),
+you, don't worry about it, you're probably not doing something that will run afoul of it).\n\n\n")
         )
         .arg(
             Arg::with_name("symbols-tmp")
+                .long("symbols-tmp")
+                .takes_value(true)
                 .long_help("A directory to use as temp space for downloading symbols.
 
 A temp dir is necessary to allow for multiple rust-minidump instances to share a cache without \
@@ -109,45 +154,32 @@ See the rust documentation for how to set that value if you wish to use somethin
 your system's default temp directory.
 
 symbols-tmp must be on the same filesystem as symbols-cache (if that doesn't mean anything to \
-you, don't worry about it, you're probably not doing something that will run afoul of it).
-\n\n\n")
-                .long("symbols-tmp")
-                .takes_value(true),
+you, don't worry about it, you're probably not doing something that will run afoul of it).\n\n\n")
         )
         .arg(
             Arg::with_name("symbol-download-timeout-secs")
                 .long("symbol-download-timeout-secs")
-                .help("The maximum amount of time (in seconds) a symbol file download is allowed \
-to take. This is necessary to enforce forward progress on misbehaving http responses.")
                 .default_value("1000")
-                .takes_value(true),
-        )
-        .arg(
-            Arg::with_name("output-file")
-                .long("output-file")
-                .help("Where to write the output to (if unspecified, stdout is used)")
-                .takes_value(true),
-        )
-        .arg(
-            Arg::with_name("log-file")
-                .long("log-file")
-                .help("Where to write logs to (if unspecified, stderr is used)")
-                .takes_value(true),
+                .takes_value(true)
+                .help("The maximum amount of time (in seconds) a symbol file download is allowed \
+to take.
+
+This is necessary to enforce forward progress on misbehaving http responses.\n\n")
         )
         .arg(
             Arg::with_name("minidump")
-                .help("Path to the minidump file to analyze.")
                 .required(true)
                 .takes_value(true)
+                .help("Path to the minidump file to analyze.")
         )
         .arg(
             Arg::with_name("symbols-path")
+                .multiple(true)
+                .takes_value(true)
                 .long_help("Path to a symbol file.
 
 If multiple symbols-path values are provided, all symbol files will be merged \
-into minidump-stackwalk's symbol database.\n\n\n")
-                .multiple(true)
-                .takes_value(true)
+into minidump-stackwalk's symbol database.\n\n\n")   
         )
         .after_help("
 
@@ -254,6 +286,14 @@ native debuginfo formats. We recommend using a version of dump_syms to generate
 
     let temp_dir = std::env::temp_dir();
 
+    let symbols_paths = matches
+        .values_of_os("symbols-path")
+        .map(|v| {
+            v.map(|os_str| Path::new(os_str).to_owned())
+                .collect::<Vec<_>>()
+        })
+        .unwrap_or_else(Vec::new);
+
     // Default to env::temp_dir()/rust-minidump-cache
     let symbols_cache = matches
         .value_of_os("symbols-cache")
@@ -277,68 +317,97 @@ native debuginfo formats. We recommend using a version of dump_syms to generate
         .map(Duration::from_secs)
         .unwrap();
 
-    let pretty = matches.is_present("pretty");
-    let human = matches.is_present("human");
-
     let minidump_path = matches.value_of_os("minidump").map(Path::new).unwrap();
-    let symbols_paths = matches
-        .values_of_os("symbols-path")
-        .map(|v| {
-            v.map(|os_str| Path::new(os_str).to_owned())
-                .collect::<Vec<_>>()
-        })
-        .unwrap_or_else(Vec::new);
 
-    if pretty && human {
+    // Determine the kind of output we're producing -- json, human, or cyborg (both).
+    // Although we have a --json argument it's mostly just there to make the documentation
+    // more clear. json output is enabled by default, and --human disables it.
+    // Mutual exclusion is enforced by an ArgGroup.
+    let mut human = matches.is_present("human");
+    let mut json = !human;
+    let cyborg = matches.value_of_os("cyborg").map(Path::new);
+
+    if cyborg.is_some() {
+        human = true;
+        json = true;
+    }
+
+    // Now check if arguments that tweak the output are valid. We can't use
+    // Arg::requires because clap doesn't understand --json being implicitly enabled.
+    let pretty = matches.is_present("pretty");
+    let brief = matches.is_present("brief");
+
+    if pretty && !json {
         error!("Humans must be hideous! (The --pretty and --human flags cannot both be set)");
         std::process::exit(1);
     }
 
+    if brief && !human {
+        error!("Robots cannot be brief! (The --brief flag is only valid for --human output (or --cyborg)");
+        std::process::exit(1);
+    }
+
     // Ok now let's do the thing!!!!
 
-    if let Ok(dump) = Minidump::read_path(minidump_path) {
-        let mut provider = MultiSymbolProvider::new();
+    match Minidump::read_path(minidump_path) {
+        Ok(dump) => {
+            let mut provider = MultiSymbolProvider::new();
 
-        if !symbols_urls.is_empty() {
-            provider.add(Box::new(Symbolizer::new(http_symbol_supplier(
-                symbols_paths,
-                symbols_urls,
-                symbols_cache,
-                symbols_tmp,
-                timeout,
-            ))));
-        } else if !symbols_paths.is_empty() {
-            provider.add(Box::new(Symbolizer::new(simple_symbol_supplier(
-                symbols_paths,
-            ))));
-        }
+            if !symbols_urls.is_empty() {
+                provider.add(Box::new(Symbolizer::new(http_symbol_supplier(
+                    symbols_paths,
+                    symbols_urls,
+                    symbols_cache,
+                    symbols_tmp,
+                    timeout,
+                ))));
+            } else if !symbols_paths.is_empty() {
+                provider.add(Box::new(Symbolizer::new(simple_symbol_supplier(
+                    symbols_paths,
+                ))));
+            }
 
-        match minidump_processor::process_minidump_with_options(&dump, &provider, options) {
-            Ok(state) => {
-                let mut stdout;
-                let mut output_f;
+            match minidump_processor::process_minidump_with_options(&dump, &provider, options) {
+                Ok(state) => {
+                    let mut stdout;
+                    let mut output_f;
+                    let cyborg_output_f = cyborg.map(|path| File::create(path).unwrap());
 
-                let mut output: &mut dyn Write = if let Some(output_path) = output_file {
-                    output_f = File::create(output_path).unwrap();
-                    &mut output_f
-                } else {
-                    stdout = std::io::stdout();
-                    &mut stdout
-                };
+                    let mut output: &mut dyn Write = if let Some(output_path) = output_file {
+                        output_f = File::create(output_path).unwrap();
+                        &mut output_f
+                    } else {
+                        stdout = std::io::stdout();
+                        &mut stdout
+                    };
 
-                if human {
-                    state.print(&mut output).unwrap();
-                } else {
-                    state.print_json(&mut output, pretty).unwrap();
+                    // Print the human output if requested (always uses the "real" output).
+                    if human {
+                        if brief {
+                            state.print_brief(&mut output).unwrap();
+                        } else {
+                            state.print(&mut output).unwrap();
+                        }
+                    }
+
+                    // Print the json output if requested (using "cyborg" output if available).
+                    if json {
+                        if let Some(mut cyborg_output_f) = cyborg_output_f {
+                            state.print_json(&mut cyborg_output_f, pretty).unwrap();
+                        } else {
+                            state.print_json(&mut output, pretty).unwrap();
+                        }
+                    }
+                }
+                Err(err) => {
+                    error!("Error processing dump: {}", err);
+                    std::process::exit(1);
                 }
             }
-            Err(err) => {
-                error!("Error processing dump: {:?}", err);
-                std::process::exit(1);
-            }
         }
-    } else {
-        error!("Error reading dump");
-        std::process::exit(1);
+        Err(err) => {
+            error!("Error reading dump: {}", err);
+            std::process::exit(1);
+        }
     }
 }
